"""
Project: BlueMath_tk
Sub-Module: wrappers
Author: GeoOcean Research Group, Universidad de Cantabria
Creation Date: 9 December 2024
License: MIT
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: Under development (NOT Working)
"""

# Import essential functions/classes to be available at the package level.
from ._base_wrappers import BaseModelWrapper
from .swash.swash_wrapper import SwashModelWrapper

# Optionally, define the module's `__all__` variable to control what gets imported when using `from module import *`.
__all__ = [
    "BaseModelWrapper",
    "SwashModelWrapper",
]



import sys
import pandas as pd
import xarray as xr
from .._base_wrappers import BaseModelWrapper


class SwashModelWrapper(BaseModelWrapper):
    """
    Wrapper for the SWASH model.
    https://swash.sourceforge.io/online_doc/swashuse/swashuse.html#input-and-output-files

    Attributes
    ----------
    swash_exec : str
        The SWASH executable path.
    default_parameters : dict
        The default parameters type for the model.

    Methods
    -------
    set_swash_exec(swash_exec: str) -> None
        Set the SWASH executable path.
    _read_tabfile(file_path: str) -> pd.DataFrame
        Read a tab file and return a pandas DataFrame.
    _convert_output_tabs_to_nc(case_id: int, output_path: str, run_path: str) -> xr.Dataset
        Convert output tabs files to a netCDF file.
    run_model(case_dir: str, log_file: str = "swash_exec.log") -> None
        Run the SWASH model for the specified case.
    """

    default_parameters = {
        "vegetation_height": float,
    }

    def __init__(
        self,
        templates_dir: str,
        templates_name: dict,
        model_parameters: dict,
        output_dir: str,
    ) -> None:
        """
        Initialize the SWASH model wrapper.

        Parameters
        ----------
        templates_dir : str
            The directory where the templates are stored.
        templates_name : list
            The names of the templates.
        model_parameters : dict
            The parameters to be used in the templates.
        output_dir : str
            The directory where the output files will be saved.
        """

        super().__init__(
            templates_dir=templates_dir,
            templates_name=templates_name,
            model_parameters=model_parameters,
            output_dir=output_dir,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(self.__class__.__name__)
        self._swash_exec: str = None

    @property
    def swash_exec(self) -> str:
        return self._swash_exec

    def set_swash_exec(self, swash_exec: str) -> None:
        self._swash_exec = swash_exec

    @staticmethod
    def _read_tabfile(file_path: str) -> pd.DataFrame:
        """
        Read a tab file and return a pandas DataFrame.
        This function is used to read the output files of SWASH.

        Parameters
        ----------
        file_path : str
            The file path.

        Returns
        -------
        pd.DataFrame
            The pandas DataFrame.
        """

        f = open(file_path, "r")
        lines = f.readlines()
        # read head colums (variables names)
        names = lines[4].split()
        names = names[1:]  # Eliminate '%'
        # read data rows
        values = pd.Series(lines[7:]).str.split(expand=True).values.astype(float)
        df = pd.DataFrame(values, columns=names)
        f.close()

        return df

    def _convert_output_tabs_to_nc(
        self, case_id: int, output_path: str, run_path: str
    ) -> xr.Dataset:
        """
        Convert tab files to a netCDF file.

        Parameters
        ----------
        output_path : str
            The output path.
        run_path : str + MDA
            The run path.

        Returns
        -------
        xr.Dataset
            The xarray Dataset.
        """

        df_output = self._read_tabfile(file_path=output_path)
        df_output.set_index(
            ["Xp", "Yp", "Tsec"], inplace=True
        )  # set index to Xp, Yp and Tsec
        ds_ouput = df_output.to_xarray()

        df_run = self._read_tabfile(file_path=run_path)
        df_run.set_index(["Tsec"], inplace=True)
        ds_run = df_run.to_xarray()

        # merge output files to one xarray.Dataset
        ds = xr.merge([ds_ouput, ds_run], compat="no_conflicts")

        # assign correct coordinate case_id
        ds.coords["case_id"] = case_id

        return ds

    def run_model(self, case_dir: str, log_file: str = "swash_exec.log") -> None:
        """
        Run the SWASH model for the specified case.

        Parameters
        ----------
        case_dir : str
            The case directory.
        log_file : str, optional
            The log file name. Default is "swash_exec.log".

        Raises
        ------
        ValueError
            If the SWASH executable was not set.
        """

        if not self.swash_exec:
            raise ValueError("The SWASH executable was not set.")
        # check if windows OS
        is_win = sys.platform.startswith("win")
        if is_win:
            cmd = "cd {0} && {1} input".format(case_dir, self.swash_exec)
        else:
            cmd = "cd {0} && {1} -input input.sws".format(case_dir, self.swash_exec)
        # redirect output
        cmd += f" 2>&1 > {log_file}"
        # execute command
        self._exec_bash_commands(str_cmd=cmd)



from typing import List
import copy
import time
import numpy as np
import pandas as pd
from scipy.optimize import fminbound
from scipy.interpolate import RBFInterpolator
from sklearn.metrics import mean_squared_error
from ._base_interpolation import BaseInterpolation
from ..core.decorators import validate_data_rbf


class RBFError(Exception):
    """
    Custom exception for RBF class.
    """

    def __init__(self, message: str = "RBF error occurred."):
        self.message = message
        super().__init__(self.message)


class RBF(BaseInterpolation):
    """
    Radial Basis Function (RBF) interpolation model.

    Here, scipy's RBFInterpolator is used to interpolate the data.
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.RBFInterpolator.html

    Attributes
    ----------
    sigma_min : float
        The minimum value for the sigma parameter.
        This value might change in the optimization process.
    sigma_max : float
        The maximum value for the sigma parameter.
        This value might change in the optimization process.
    sigma_diff : float
        The minimum difference between the optimal sigma and the minimum and maximum
        sigma values.
    kernel : str
        Type of RBF. This should be one of

            - 'linear'               : ``-r``
            - 'thin_plate_spline'    : ``r**2 * log(r)``
            - 'cubic'                : ``r**3``
            - 'quintic'              : ``-r**5``
            - 'multiquadric'         : ``-sqrt(1 + r**2)``
            - 'inverse_multiquadric' : ``1/sqrt(1 + r**2)``
            - 'inverse_quadratic'    : ``1/(1 + r**2)``
            - 'gaussian'             : ``exp(-r**2)``

    smoothing : float or (npoints, ) array_like
        Smoothing parameter. The interpolant perfectly fits the data when this
        is set to 0. For large values, the interpolant approaches a least
        squares fit of a polynomial with the specified degree.
    degree : int
        Degree of the added polynomial. For some RBFs the interpolant may not
        be well-posed if the polynomial degree is too small. Those RBFs and
        their corresponding minimum degrees are

            - 'multiquadric'      : 0
            - 'linear'            : 0
            - 'thin_plate_spline' : 1
            - 'cubic'             : 1
            - 'quintic'           : 2

        The default value is the minimum degree for `kernel` or 0 if there is
        no minimum degree. Set this to -1 for no added polynomial.
    neighbors : int
        If specified, the value of the interpolant at each evaluation point
        will be computed using only this many nearest data points. All the data
        points are used by default.
    rbfs : dict
        Dict with RBFInterpolator instances.
    subset_data : pd.DataFrame
        The subset data used to fit the model.
    normalized_subset_data : pd.DataFrame
        The normalized subset data used to fit the model.
    target_data : pd.DataFrame
        The target data used to fit the model.
    normalized_target_data : pd.DataFrame
        The normalized target data used to fit the model.
        This attribute is only set if normalize_target_data is True in the fit method.
    subset_directional_variables : List[str]
        The subset directional variables.
    target_directional_variables : List[str]
        The target directional variables.
    subset_processed_variables : List[str]
        The subset processed variables.
    target_processed_variables : List[str]
        The target processed variables.
    subset_custom_scale_factor : dict
        The custom scale factor for the subset data.
    target_custom_scale_factor : dict
        The custom scale factor for the target data.
    subset_scale_factor : dict
        The scale factor for the subset data.
    target_scale_factor : dict
        The scale factor for the target data.
    rbf_coeffs : pd.DataFrame
        The RBF coefficients for the target variables.
    opt_sigmas : dict
        The optimal sigmas for the target variables.

    Methods
    -------
    fit : Fits the model to the data.
    predict : Predicts the data for the provided dataset.
    fit_predict : Fits the model to the subset and predicts the interpolated dataset.

    See Also
    --------
    OtherInterpolationMethod

    References
    ----------
    .. [1] Fasshauer, G., 2007. Meshfree Approximation Methods with Matlab.
        World Scientific Publishing Co.

    .. [2] http://amadeus.math.iit.edu/~fass/603_ch3.pdf

    .. [3] Wahba, G., 1990. Spline Models for Observational Data. SIAM.

    .. [4] http://pages.stat.wisc.edu/~wahba/stat860public/lect/lect8/lect8.pdf

    Examples
    --------
    >>> import numpy as np

    Notes
    -----
    .. versionadded:: 1.0.3
    TODO: For the moment, this class only supports optimization for one parameter kernels.
          For this reason, we only have sigma as the parameter to optimize.
          This sigma refers to the sigma parameter in the Gaussian kernel (but is used for all kernels).
    """

    def __init__(
        self,
        sigma_min: float = 0.001,
        sigma_max: float = 0.1,
        sigma_diff: float = 0.0001,
        kernel: str = "gaussian",
        smoothing: float = 0.0,
        degree: int = None,
        neighbors: int = None,
    ):
        """
        Initializes the RBF model.

        Parameters
        ----------
        sigma_min : float, optional
            The minimum value for the sigma parameter. Default is 0.001.
        sigma_max : float, optional
            The maximum value for the sigma parameter. Default is 0.1.
        sigma_diff : float, optional
            The minimum difference between the optimal sigma and the minimum and maximum
            sigma values. Default is 0.0001.
        kernel : str, optional
            Type of RBF. Default is 'gaussian'.
        smoothing : float, optional
            Smoothing parameter. Default is 0.0.
        degree : int, optional
            Degree of the added polynomial. Default is None.
        neighbors : int, optional
            If specified, the value of the interpolant at each evaluation point will be
            computed using only this many nearest data points. Default is None.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)
        if not isinstance(sigma_min, float) or sigma_min < 0:
            raise ValueError("sigma_min must be a positive float.")
        self._sigma_min = sigma_min
        if not isinstance(sigma_max, float) or sigma_max < sigma_min:
            raise ValueError(
                "sigma_max must be a positive float greater than sigma_min."
            )
        self._sigma_max = sigma_max
        if not isinstance(sigma_diff, float) or sigma_diff < 0:
            raise ValueError("sigma_diff must be a positive float.")
        self._sigma_diff = sigma_diff
        if not isinstance(kernel, str):
            raise ValueError("kernel must be a string.")
        self._kernel = kernel
        if not isinstance(smoothing, float):
            raise ValueError("smoothing must be a float.")
        self._smoothing = smoothing
        if not isinstance(degree, int) and degree is not None:
            raise ValueError("degree must be an integer.")
        self._degree = degree
        if not isinstance(neighbors, int) and neighbors is not None:
            raise ValueError("neighbors must be an integer.")
        self._neighbors = neighbors
        self._rbfs: dict = {}  # Dict with RBFInterpolator instances
        # Below, we initialize the attributes that will be set in the fit method
        self.is_fitted: bool = False
        self.is_target_normalized: bool = False
        self._subset_data: pd.DataFrame = pd.DataFrame()
        self._normalized_subset_data: pd.DataFrame = pd.DataFrame()
        self._target_data: pd.DataFrame = pd.DataFrame()
        self._normalized_target_data: pd.DataFrame = pd.DataFrame()
        self._subset_directional_variables: List[str] = []
        self._target_directional_variables: List[str] = []
        self._subset_processed_variables: List[str] = []
        self._target_processed_variables: List[str] = []
        self._subset_custom_scale_factor: dict = {}
        self._target_custom_scale_factor: dict = {}
        self._subset_scale_factor: dict = {}
        self._target_scale_factor: dict = {}
        self._rbf_coeffs: pd.DataFrame = pd.DataFrame()
        self._opt_sigmas: dict = {}

    @property
    def sigma_min(self) -> float:
        return self._sigma_min

    @property
    def sigma_max(self) -> float:
        return self._sigma_max

    @property
    def sigma_diff(self) -> float:
        return self._sigma_diff

    @property
    def kernel(self) -> str:
        return self._kernel

    @property
    def smoothing(self) -> float:
        return self._smoothing

    @property
    def degree(self) -> int:
        return self._degree

    @property
    def neighbors(self) -> int:
        return self._neighbors

    @property
    def rbfs(self) -> dict:
        return self._rbfs

    @property
    def subset_data(self) -> pd.DataFrame:
        return self._subset_data

    @property
    def normalized_subset_data(self) -> pd.DataFrame:
        return self._normalized_subset_data

    @property
    def target_data(self) -> pd.DataFrame:
        return self._target_data

    @property
    def normalized_target_data(self) -> pd.DataFrame:
        if self._normalized_target_data.empty:
            raise ValueError("Target data is not normalized.")
        return self._normalized_target_data

    @property
    def subset_directional_variables(self) -> List[str]:
        return self._subset_directional_variables

    @property
    def target_directional_variables(self) -> List[str]:
        return self._target_directional_variables

    @property
    def subset_processed_variables(self) -> List[str]:
        return self._subset_processed_variables

    @property
    def target_processed_variables(self) -> List[str]:
        return self._target_processed_variables

    @property
    def subset_custom_scale_factor(self) -> dict:
        return self._subset_custom_scale_factor

    @property
    def target_custom_scale_factor(self) -> dict:
        return self._target_custom_scale_factor

    @property
    def subset_scale_factor(self) -> dict:
        return self._subset_scale_factor

    @property
    def target_scale_factor(self) -> dict:
        return self._target_scale_factor

    @property
    def rbf_coeffs(self) -> pd.DataFrame:
        return self._rbf_coeffs

    @property
    def opt_sigmas(self) -> dict:
        if not self._opt_sigmas:
            raise ValueError("Specified kernel does not require optimization.")
        return self._opt_sigmas

    def _preprocess_subset_data(
        self, subset_data: pd.DataFrame, is_fit: bool = True
    ) -> pd.DataFrame:
        """
        This function preprocesses the subset data.

        Parameters
        ----------
        subset_data : pd.DataFrame
            The subset data to preprocess (could be a dataset to predict).
        is_fit : bool, optional
            Whether the data is to fit or not. Default is True.

        Returns
        -------
        pd.DataFrame
            The preprocessed subset data.

        Raises
        ------
        ValueError
            If the subset contains NaNs.

        Notes
        -----
        - This function preprocesses the subset data by:
            - Checking for NaNs.
            - Preprocessing directional variables.
            - Normalizing the data.
        """

        # Make copies to avoid modifying the original data
        subset_data = subset_data.copy()

        self.logger.info("Checking for NaNs in subset data")
        subset_data = self.check_nans(data=subset_data, raise_error=True)

        self.logger.info("Preprocessing subset data")
        for directional_variable in self.subset_directional_variables:
            var_u_component, var_y_component = self.get_uv_components(
                x_deg=subset_data[directional_variable].values
            )
            subset_data[f"{directional_variable}_u"] = var_u_component
            subset_data[f"{directional_variable}_v"] = var_y_component
            # Drop the original directional variable in subset_data
            subset_data.drop(columns=[directional_variable], inplace=True)
        self._subset_processed_variables = list(subset_data.columns)

        self.logger.info("Normalizing subset data")
        normalized_subset_data, subset_scale_factor = self.normalize(
            data=subset_data,
            custom_scale_factor=self.subset_custom_scale_factor
            if is_fit
            else self.subset_scale_factor,
        )

        self.logger.info("Subset data preprocessed successfully")

        if is_fit:
            self._subset_data = subset_data
            self._normalized_subset_data = normalized_subset_data
            self._subset_scale_factor = subset_scale_factor

        return normalized_subset_data.copy()

    def _preprocess_target_data(
        self,
        target_data: pd.DataFrame,
        normalize_target_data: bool = True,
    ) -> pd.DataFrame:
        """
        This function preprocesses the target data.

        Parameters
        ----------
        target_data : pd.DataFrame
            The target data to preprocess.
        normalize_target_data : bool, optional
            Whether to normalize the target data. Default is True.

        Returns
        -------
        pd.DataFrame
            The preprocessed target data.

        Raises
        ------
        ValueError
            If the target contains NaNs.

        Notes
        -----
        - This function preprocesses the target data by:
            - Checking for NaNs.
            - Preprocessing directional variables.
            - Normalizing the data.
        """

        # Make copies to avoid modifying the original data
        target_data = target_data.copy()

        self.logger.info("Checking for NaNs in target data")
        target_data = self.check_nans(data=target_data, raise_error=True)

        self.logger.info("Preprocessing target data")
        for directional_variable in self.target_directional_variables:
            var_u_component, var_y_component = self.get_uv_components(
                x_deg=target_data[directional_variable].values
            )
            target_data[f"{directional_variable}_u"] = var_u_component
            target_data[f"{directional_variable}_v"] = var_y_component
            # Drop the original directional variable in target_data
            target_data.drop(columns=[directional_variable], inplace=True)
        self._target_processed_variables = list(target_data.columns)

        if normalize_target_data:
            self.logger.info("Normalizing target data")
            normalized_target_data, target_scale_factor = self.normalize(
                data=target_data,
                custom_scale_factor=self.target_custom_scale_factor,
            )
            self.is_target_normalized = True
            self._target_data = target_data.copy()
            self._normalized_target_data = normalized_target_data.copy()
            self._target_scale_factor = target_scale_factor.copy()
            self.logger.info("Target data preprocessed successfully")
            return normalized_target_data.copy()

        else:
            self.is_target_normalized = False
            self._target_data = target_data.copy()
            self._normalized_target_data = pd.DataFrame()
            self._target_scale_factor = {}
            self.logger.info("Target data preprocessed successfully")
            return target_data.copy()

    def _cost_sigma(self, sigma: float, x: np.ndarray, y: np.ndarray) -> float:
        """
        This function is called by fminbound to minimize the cost function.

        Parameters
        ----------
        sigma : float
            The sigma parameter for the kernel.
        x : np.ndarray
            The subset data used to interpolate.
        y : np.ndarray
            The target data to interpolate.

        Returns
        -------
        float
            The cost value.
        """

        # Instantiate the RBFInterpolator
        rbf = RBFInterpolator(
            y=x,
            d=y,
            neighbors=self.neighbors,
            smoothing=self.smoothing,
            kernel=self.kernel,
            epsilon=sigma,
            degree=self.degree,
        )
        predicted_y = rbf(x)
        # TODO: Check if this is the correct cost function
        # cost = mean_squared_error(y, predicted_y)
        cost = np.dot(y - predicted_y, y - predicted_y)

        return cost

    def _calc_opt_sigma(
        self, target_variable: np.ndarray, subset_variables: np.ndarray
    ) -> RBFInterpolator:
        """
        This function calculates the optimal sigma for the given target variable.

        Parameters
        ----------
        target_variable : np.ndarray
            The target variable to interpolate.
        subset_variables : np.ndarray
            The subset variables used to interpolate.

        Returns
        -------
        float
            The optimal sigma.
        """

        t0 = time.time()
        # Initialize sigma_min, sigma_max, and d_sigma
        sigma_min, sigma_max, d_sigma = self.sigma_min, self.sigma_max, 0

        # Loop until sigma_diff is less than the specified sigma_diff
        while d_sigma < self.sigma_diff:
            opt_sigma = fminbound(
                func=self._cost_sigma,
                x1=sigma_min,
                x2=sigma_max,
                args=(subset_variables, target_variable),
                disp=0,
            )
            lm_min = np.abs(opt_sigma - sigma_min)
            lm_max = np.abs(opt_sigma - sigma_max)
            if lm_min < self.sigma_diff:
                sigma_min = sigma_min - sigma_min / 2
            elif lm_max < self.sigma_min:
                sigma_max = sigma_max + sigma_max / 2
            d_sigma = np.nanmin([lm_min, lm_max])

        # Save the fitted RBF for the optimal sigma
        rbf = RBFInterpolator(
            y=subset_variables,
            d=target_variable,
            neighbors=self.neighbors,
            smoothing=self.smoothing,
            kernel=self.kernel,
            epsilon=opt_sigma,
            degree=self.degree,
        )

        # Calculate the time taken to optimize sigma
        t1 = time.time()
        self.logger.info(f"Optimal sigma: {opt_sigma} - Time: {t1 - t0:.2f} seconds")

        return rbf, opt_sigma

    @validate_data_rbf
    def fit(
        self,
        subset_data: pd.DataFrame,
        target_data: pd.DataFrame,
        subset_directional_variables: List[str] = [],
        target_directional_variables: List[str] = [],
        subset_custom_scale_factor: dict = {},
        normalize_target_data: bool = True,
        target_custom_scale_factor: dict = {},
        num_threads: int = None,
    ) -> None:
        """
        Fits the model to the data.

        Parameters
        ----------
        subset_data : pd.DataFrame
            The subset data used to fit the model.
        target_data : pd.DataFrame
            The target data used to fit the model.
        subset_directional_variables : List[str], optional
            The subset directional variables. Default is [].
        target_directional_variables : List[str], optional
            The target directional variables. Default is [].
        subset_custom_scale_factor : dict, optional
            The custom scale factor for the subset data. Default is {}.
        normalize_target_data : bool, optional
            Whether to normalize the target data. Default is True.
        target_custom_scale_factor : dict, optional
            The custom scale factor for the target data. Default is {}.
        num_threads : int, optional
            The number of threads to use for the optimization. Default is None.

        Notes
        -----
        - This function fits the RBF model to the data by:
            1. Preprocessing the subset and target data.
            2. Calculating the optimal sigma for the target variables.
            3. Storing the RBF coefficients and optimal sigmas.
        - The number of threads to use for the optimization can be specified.
        """

        if num_threads is not None:
            self.set_num_processors_to_use(num_processors=num_threads)
            self.logger.info(f"Using {num_threads} threads for optimization.")

        self._subset_directional_variables = subset_directional_variables
        self._target_directional_variables = target_directional_variables
        self._subset_custom_scale_factor = subset_custom_scale_factor
        self._target_custom_scale_factor = target_custom_scale_factor
        subset_data = self._preprocess_subset_data(subset_data=subset_data)
        target_data = self._preprocess_target_data(
            target_data=target_data,
            normalize_target_data=normalize_target_data,
        )

        self.logger.info("Fitting RBF model to the data")
        # RBF fitting for all variables
        rbf_coeffs, opt_sigmas = {}, {}

        # Optimize sigma for each target variable
        for target_var in target_data.columns:
            self.logger.info(f"Fitting RBF for variable {target_var}")
            target_var_values = target_data[target_var].values
            if (
                self.kernel == "linear"
                or self.kernel == "cubic"
                or self.kernel == "quintic"
                or self.kernel == "thin_plate_spline"
            ):
                rbf = RBFInterpolator(
                    y=subset_data.values,
                    d=target_var_values,
                    neighbors=self.neighbors,
                    smoothing=self.smoothing,
                    kernel=self.kernel,
                    degree=self.degree,
                )
                opt_sigma = None
            else:
                rbf, opt_sigma = self._calc_opt_sigma(
                    target_variable=target_var_values,
                    subset_variables=subset_data.values,
                )
            self.rbfs[target_var] = copy.deepcopy(rbf)
            rbf_coeffs[target_var] = rbf._coeffs.flatten()
            opt_sigmas[target_var] = opt_sigma

        # Store the RBF coefficients and optimal sigmas
        self._rbf_coeffs = pd.DataFrame(rbf_coeffs)
        self._opt_sigmas = opt_sigmas

        # Set the is_fitted attribute to True
        self.is_fitted = True

    def predict(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """
        Predicts the data for the provided dataset.

        Parameters
        ----------
        dataset : pd.DataFrame
            The dataset to predict (must have same variables than subset).

        Returns
        -------
        pd.DataFrame
            The interpolated dataset.

        Raises
        ------
        ValueError
            If the model is not fitted.

        Notes
        -----
        - This function predicts the data by:
            1. Reconstructing the data using the fitted coefficients.
            2. Denormalizing the target data if normalize_target_data is True.
            3. Calculating the degrees for the target directional variables.
        """

        if self.is_fitted is False:
            raise RBFError("RBF model must be fitted before predicting.")

        self.logger.info("Reconstructing data using fitted coefficients.")
        normalized_dataset = self._preprocess_subset_data(
            subset_data=dataset, is_fit=False
        )

        # Create an empty array to store the interpolated target data
        interpolated_target_array = np.zeros(
            (normalized_dataset.shape[0], len(self.target_processed_variables))
        )
        for target_var in self.target_processed_variables:
            self.logger.info(f"Predicting target variable {target_var}")
            rbf = self.rbfs[target_var]
            interpolated_target_array[
                :, self.target_processed_variables.index(target_var)
            ] = rbf(normalized_dataset.values)
        interpolated_target = pd.DataFrame(
            data=interpolated_target_array, columns=self.target_processed_variables
        )

        # Denormalize the target data if normalize_target_data is True
        if self.is_target_normalized:
            self.logger.info("Denormalizing target data")
            interpolated_target = self.denormalize(
                normalized_data=interpolated_target,
                scale_factor=self.target_scale_factor,
            )

        # Calculate the degrees for the target directional variables
        for directional_variable in self.target_directional_variables:
            self.logger.info(f"Calculating target degrees for {directional_variable}")
            interpolated_target[directional_variable] = self.get_degrees_from_uv(
                xu=interpolated_target[f"{directional_variable}_u"].values,
                xv=interpolated_target[f"{directional_variable}_v"].values,
            )

        return interpolated_target

    def fit_predict(
        self,
        subset_data: pd.DataFrame,
        target_data: pd.DataFrame,
        dataset: pd.DataFrame,
        subset_directional_variables: List[str] = [],
        target_directional_variables: List[str] = [],
        subset_custom_scale_factor: dict = {},
        normalize_target_data: bool = True,
        target_custom_scale_factor: dict = {},
        num_threads: int = None,
    ) -> pd.DataFrame:
        """
        Fits the model to the subset and predicts the interpolated dataset.

        Parameters
        ----------
        subset_data : pd.DataFrame
            The subset data used to fit the model.
        target_data : pd.DataFrame
            The target data used to fit the model.
        dataset : pd.DataFrame
            The dataset to predict (must have same variables than subset).
        subset_directional_variables : List[str], optional
            The subset directional variables. Default is [].
        target_directional_variables : List[str], optional
            The target directional variables. Default is [].
        subset_custom_scale_factor : dict, optional
            The custom scale factor for the subset data. Default is {}.
        normalize_target_data : bool, optional
            Whether to normalize the target data. Default is True.
        target_custom_scale_factor : dict, optional
            The custom scale factor for the target data. Default is {}.
        num_threads : int, optional
            The number of threads to use for the optimization. Default is None.

        Returns
        -------
        pd.DataFrame
            The interpolated dataset.

        Notes
        -----
        - This function fits the model to the subset and predicts the interpolated dataset.
        """

        self.fit(
            subset_data=subset_data,
            target_data=target_data,
            subset_directional_variables=subset_directional_variables,
            target_directional_variables=target_directional_variables,
            subset_custom_scale_factor=subset_custom_scale_factor,
            normalize_target_data=normalize_target_data,
            target_custom_scale_factor=target_custom_scale_factor,
            num_threads=num_threads,
        )

        return self.predict(dataset=dataset)



import keras
from keras import layers
from typing import List, Tuple


def ResidualBlock(width: int) -> layers.Layer:
    def apply(x: layers.Layer) -> layers.Layer:
        input_width = x.shape[3]
        residual = x if input_width == width else layers.Conv2D(width, kernel_size=1)(x)

        x = layers.LayerNormalization(axis=-1, center=True, scale=True)(x)
        x = layers.Conv2D(
            width, kernel_size=3, padding="same", activation=keras.activations.swish
        )(x)
        x = layers.Conv2D(width, kernel_size=3, padding="same")(x)
        x = layers.Add()([x, residual])
        return x

    return apply


def DownBlock(width: int, block_depth: int) -> layers.Layer:
    def apply(
        x: Tuple[layers.Layer, List[layers.Layer]],
    ) -> Tuple[layers.Layer, List[layers.Layer]]:
        x, skips = x
        for _ in range(block_depth):
            x = ResidualBlock(width)(x)
            skips.append(x)
        x = layers.AveragePooling2D(pool_size=2)(x)
        return x, skips

    return apply


def UpBlock(width: int, block_depth: int) -> layers.Layer:
    def apply(x: Tuple[layers.Layer, List[layers.Layer]]) -> layers.Layer:
        x, skips = x
        x = layers.UpSampling2D(size=2, interpolation="bilinear")(x)
        for _ in range(block_depth):
            x = layers.Concatenate()([x, skips.pop()])
            x = ResidualBlock(width)(x)
        return x

    return apply


def get_model(
    image_height: int,
    image_width: int,
    input_frames: int,
    output_frames: int,
    down_widths: List[int] = [64, 128, 256],
    up_widths: List[int] = [256, 128, 64],
    block_depth: int = 2,
) -> keras.Model:
    """Builds the U-Net like model with residual blocks and skip connections."""

    inputs = keras.Input(shape=(image_height, image_width, input_frames))
    x = layers.Conv2D(down_widths[0], kernel_size=1)(inputs)

    skips = []
    for width in down_widths[:-1]:
        x, skips = DownBlock(width, block_depth)([x, skips])

    for _ in range(block_depth):
        x = ResidualBlock(down_widths[-1])(x)

    for width in up_widths[1:]:
        x = UpBlock(width, block_depth)([x, skips])

    outputs = layers.Conv2D(output_frames, kernel_size=1, kernel_initializer="zeros")(x)

    return keras.Model(inputs, outputs, name="residual_unet")



import unittest
import os
import pandas as pd
from bluemath_tk.datamining.lhs import LHS


class TestLHS(unittest.TestCase):
    def setUp(self):
        self.dimensions_names = ["CM", "SS", "Qb"]
        self.lower_bounds = [0.5, -0.2, 1]
        self.upper_bounds = [5.3, 1.5, 200]
        self.lhs = LHS(num_dimensions=3, seed=0)

    def test_generate(self):
        lhs_sampled_df = self.lhs.generate(
            dimensions_names=self.dimensions_names,
            lower_bounds=self.lower_bounds,
            upper_bounds=self.upper_bounds,
            num_samples=100,
        )
        self.assertIsInstance(lhs_sampled_df, pd.DataFrame)
        self.assertEqual(lhs_sampled_df.shape[0], 100)

    # def test_save_model(self):
    #     self.lhs.save_model("/tmp/lhs_model.pkl")
    #     self.assertTrue(os.path.exists("/tmp/lhs_model.pkl"))

    # def test_load_model(self):
    #     lhs_loaded = self.lhs.load_model("/tmp/lhs_model.pkl")
    #     self.assertIsInstance(lhs_loaded, LHS)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import pandas as pd
from bluemath_tk.interpolation.rbf import RBF


class TestRBF(unittest.TestCase):
    def setUp(self):
        self.dataset = pd.DataFrame(
            {
                "Hs": np.random.rand(1000) * 7,
                "Tp": np.random.rand(1000) * 20,
                "Dir": np.random.rand(1000) * 360,
            }
        )
        self.subset = self.dataset.sample(frac=0.25)
        self.target = pd.DataFrame(
            {
                "HsPred": self.subset["Hs"] * 2 + self.subset["Tp"] * 3,
                "DirPred": -self.subset["Dir"],
            }
        )
        self.rbf = RBF()

    def test_fit(self):
        self.rbf.fit(
            subset_data=self.subset,
            subset_directional_variables=["Dir"],
            target_data=self.target,
            target_directional_variables=["DirPred"],
            normalize_target_data=True,
        )
        self.assertTrue(self.rbf.is_fitted)
        self.assertTrue(self.rbf.is_target_normalized)
        self.assertIn("Dir_u", self.rbf.normalized_subset_data.columns)
        self.assertIn("Dir_v", self.rbf.normalized_subset_data.columns)
        self.assertIn("DirPred_u", self.rbf.normalized_target_data.columns)
        self.assertIn("DirPred_v", self.rbf.normalized_target_data.columns)
        self.assertFalse(self.rbf.rbf_coeffs.empty)
        self.assertFalse(self.rbf.opt_sigmas == {})

    def test_predict(self):
        self.rbf.fit(
            subset_data=self.subset,
            subset_directional_variables=["Dir"],
            target_data=self.target,
            target_directional_variables=["DirPred"],
            normalize_target_data=True,
        )
        predictions = self.rbf.predict(dataset=self.dataset)
        self.assertIsInstance(predictions, pd.DataFrame)
        self.assertIn("HsPred", predictions.columns)
        self.assertIn("DirPred", predictions.columns)

    def test_fit_predict(self):
        predictions = self.rbf.fit_predict(
            subset_data=self.subset,
            subset_directional_variables=["Dir"],
            target_data=self.target,
            target_directional_variables=["DirPred"],
            normalize_target_data=True,
            dataset=self.dataset,
            num_threads=4,
        )
        self.assertIsInstance(predictions, pd.DataFrame)
        self.assertIn("HsPred", predictions.columns)
        self.assertIn("DirPred", predictions.columns)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from scipy import interpolate

# custom 2D profiles for SWASH numerical model cases


def reef(dx, h0, Slope1, Slope2, Wreef, Wfore, bCrest, emsl):
    """
    Reef morphologic profile (Pearson et al. 2017)

    dx:      bathymetry mesh resolution at x axes (m)
    h0:      offshore depth (m)
    Slope1:  fore shore slope
    Slope2:  inner shore slope
    Wreef:   reef bed width (m)
    Wfore:   flume length before fore toe (m)
    bCrest:  beach heigh (m)
    emsl:    mean sea level (m)

    return depth data values
    """

    # flume length
    W_inner = bCrest / Slope2
    W1 = int(abs(h0 - emsl) / Slope1)

    # sections length
    x1 = np.arange(0, Wfore, dx)
    x2 = np.arange(0, W1, dx)
    x3 = np.arange(0, Wreef, dx)
    x4 = np.arange(0, W_inner, dx)

    # curve equation
    y_fore = np.zeros(len(x1)) + [h0]
    y1 = -Slope1 * x2 + h0
    y2 = np.zeros(len(x3)) + emsl
    y_inner = -Slope2 * x4 + emsl

    # overtopping cases: an inshore plane beach to dissipate overtopped flux
    plane = 0.005 * np.arange(0, 150, 1) + y_inner[-1]

    # concatenate depth
    depth = np.concatenate([y_fore, y1, y2, y_inner, plane])

    return depth


def linear(dx, h0, bCrest, m, Wfore):
    """
    simple linear profile (y = m * x + n)

    dx:      bathymetry mesh resolution at x axes (m)
    h0:      offshore depth (m)
    bCrest:  beach heigh (m)
    m:       profile slope
    Wfore:   flume length before slope toe (m)

    return depth data values
    """

    # Flume length
    W1 = int(h0 / m)
    W2 = int(bCrest / m)

    # Sections length
    x1 = np.arange(0, Wfore, dx)
    x2 = np.arange(0, W1, dx)
    x3 = np.arange(0, W2, dx)

    # Curve equation
    y_fore = np.zeros(len(x1)) + [h0]
    y1 = -m * x2 + h0
    y2 = -m * x3

    # Overtopping cases: an inshore plane beach to dissipate overtopped flux
    plane = 0.005 * np.arange(0, len(y2), 1) + y2[-1]  # Length bed = 2 L

    # concatenate depth
    depth = np.concatenate([y_fore, y1, y2, plane])

    return depth


def parabolic(dx, h0, A, xBeach, bCrest):
    """
    Parabolic profile (y = A * x^(2/3))

    dx:      bathymetry mesh resolution at x axes (m)
    h0:      offshore depth (m)
    A:       parabola coefficient
    xBeach:  beach length(m)
    bCrest:  beach heigh (m)
    """

    lx = np.arange(1, xBeach, dx)
    y = -(bCrest / xBeach) * lx

    depth, xl = [], []
    x, z = 0, 0

    while z <= h0:
        z = A * x ** (2 / 3)
        depth.append(z)
        xl.append(x)
        x += dx

    f = interpolate.interp1d(xl, depth)
    xnew = np.arange(0, int(np.round(len(depth) * dx)), 1)
    ynew = f(xnew)

    # concatenate depth
    depth = np.concatenate([ynew[::-1], y])

    return depth


def biparabolic(h0, hsig, omega_surf_list, TR):
    """
    Biparabolic profile (Bernabeu et al. 2013)

    h0:          offshore water level (m)
    hsig:        significant wave height (m)
    omega_surf:  intertidal dimensionless fall velocity (1 <= omega_surf <= 5)
    TR:          tidal range (m)
    """

    # Discontinuity point
    hr = 1.1 * hsig + TR

    # Legal point
    ha = 3 * hsig + TR

    # Empirical adjusted parameters
    A = 0.21 - 0.02 * omega_surf_list
    B = 0.89 * np.exp(-1.24 * omega_surf_list)
    C = 0.06 + 0.04 * omega_surf_list
    D = 0.22 * np.exp(-0.83 * omega_surf_list)

    # Different values for the height
    h = np.linspace(0, h0, 150)
    h_cont = []

    # Important points for the profile
    xr = (hr / A) ** (3 / 2) + (B / (A ** (3 / 2))) * hr**3

    # Lines for the profile
    x, Xx, X, xX = [], [], [], []

    for hs in h:  # For each vertical point
        if hs < hr:
            x_max = 0
            xapp = (hs / A) ** (3 / 2) + (B / (A ** (3 / 2))) * hs**3
            x.append(xapp)
            x_max = max(xapp, x_max)
            if hs > (hr - 1.5):
                Xxapp = (hs / C) ** (3 / 2) + (D / (C ** (3 / 2))) * hs**3
                Xx.append(Xxapp)
                h_cont.append(hs)
        else:
            Xapp = (hs / C) ** (3 / 2) + (D / (C ** (3 / 2))) * hs**3
            if (hs - hr) < 0.1:
                x_diff = x_max - Xapp
            X.append(Xapp)
            if hs < (hr + 1.5):
                xXapp = (hs / A) ** (3 / 2) + (B / (A ** (3 / 2))) * hs**3
                xX.append(xXapp)
                h_cont.append(hs)

    h_cont = np.array(h_cont)
    x_tot = np.concatenate((np.array(x), np.array(X) + x_diff))
    # x_cont = np.concatenate((np.array(Xx)+x_diff, np.array(xX)))

    # Centering the y-axis in the mean tide
    xnew = np.arange(0, x_tot[-1], 1)
    # xnew_border = np.arange(x_tot[-1]-x_cont[0], x_cont[-1]-x_cont[-1], 1)
    depth = h - TR / 2
    # border = (-h_cont+TR/2)

    f = interpolate.interp1d(x_tot, depth)
    # f1 = interpolate.interp1d(x_cont, border)
    ynew = f(xnew)[::-1]
    # ynew_border = f1(xnew_border)[::-1]

    depth = (h - TR / 2)[::-1]
    # border = (-h_cont+TR/2)[::-1]

    # plot
    # TODO: move plot to plots.py
    # fig, ax = plt.subplots(1, figsize = (12, 4))
    # ax.plot(xnew, -ynew, color='k', zorder=3)
    # ax.fill_between(xnew, np.zeros((len(xnew)))+(-ynew[0]),
    #                -ynew,facecolor="wheat", alpha=1, zorder=2)
    # ax.scatter(x_tot[-1]-xr, -hr+TR/2, s=30, c='red', label='Discontinuity point', zorder=5)
    # ax.fill_between(xnew, -ynew, np.zeros(len(xnew)), facecolor="deepskyblue", alpha=0.5, zorder=1)
    # ax.axhline(-ha+TR/2, color='grey', ls='-.', label='Available region')
    # ax.axhline(TR/2, color='silver', ls='--', label='HT')
    # ax.axhline(0, color='lightgrey', ls='--', label='MSL')
    # ax.axhline(-TR/2, color='silver', ls='--', label='LT')
    # ax.scatter(xnew_border, -ynew_border, c='k', s=1, marker='_', zorder=4)

    # attrbs
    # ax.set_ylim(-ynew[0], -ynew[-1]+1)
    # ax.set_xlim(0, x_tot[-1])
    # set_title  = '$\Omega_{sf}$ = ' + str(omega_surf_list)
    # set_title += ', TR = ' + str(TR)
    # ax.set_title(set_title)
    # ax.legend(loc='upper left')
    # ax.set_ylabel('$Depth$ $[m]$', fontweight='bold')
    # ax.set_xlabel('$X$ $[m]$', fontweight='bold')

    # TODO: deph or ynew ?
    return ynew


def custom_profile(dx, emsl, xs, ys):
    """
    custom N points profile

    dx:   bathymetry mesh resolution at x axes (m)
    xs:    x values array
    ys:    y values array
    emsl:  mean sea level (m)
    """

    # flume length
    xnew = np.arange(xs[0], xs[-1], dx)
    f = interpolate.interp1d(xs, ys)
    ynew = f(xnew)

    depth = -ynew

    return depth



import unittest
import numpy as np
import pandas as pd
from bluemath_tk.datamining.som import SOM


class TestSOM(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                "Hs": np.random.rand(1000) * 7,
                "Tp": np.random.rand(1000) * 20,
                "Dir": np.random.rand(1000) * 360,
            }
        )
        self.som = SOM(som_shape=(3, 3), num_dimensions=4)

    def test_fit(self):
        self.som.fit(data=self.df, directional_variables=["Dir"])
        self.assertIsInstance(self.som.centroids, pd.DataFrame)
        self.assertEqual(self.som.centroids.shape[0], 9)

    def test_predict(self):
        data_sample = pd.DataFrame(
            {
                "Hs": np.random.rand(15) * 7,
                "Tp": np.random.rand(15) * 20,
                "Dir": np.random.rand(15) * 360,
            }
        )
        self.som.fit(data=self.df, directional_variables=["Dir"])
        nearest_centroids, nearest_centroid_df = self.som.predict(data=data_sample)
        self.assertIsInstance(nearest_centroids, np.ndarray)
        self.assertEqual(len(nearest_centroids), 15)
        self.assertIsInstance(nearest_centroid_df, pd.DataFrame)
        self.assertEqual(nearest_centroid_df.shape[0], 15)

    def test_fit_predict(self):
        nearest_centroids, nearest_centroid_df = self.som.fit_predict(
            data=self.df, directional_variables=["Dir"]
        )
        self.assertIsInstance(nearest_centroids, np.ndarray)
        self.assertEqual(len(nearest_centroids), 1000)
        self.assertIsInstance(nearest_centroid_df, pd.DataFrame)
        self.assertEqual(nearest_centroid_df.shape[0], 1000)

    def test_activation_response(self):
        self.som.fit(data=self.df, directional_variables=["Dir"])
        act_resp = self.som.activation_response()
        self.assertIsInstance(act_resp, np.ndarray)
        self.assertEqual(act_resp.shape, (3, 3))

    def test_get_centroids_probs_for_labels(self):
        data_sample = pd.DataFrame(
            {
                "Hs": np.random.rand(15) * 7,
                "Tp": np.random.rand(15) * 20,
                "Dir": np.random.rand(15) * 360,
            }
        )
        self.som.fit(data=self.df, directional_variables=["Dir"])
        centroids_probs = self.som.get_centroids_probs_for_labels(
            data=data_sample,
            labels=np.random.randint(0, 3, 15),
        )
        self.assertIsInstance(centroids_probs, pd.DataFrame)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import xarray as xr
from typing import Union, List
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA as PCA_, IncrementalPCA as IncrementalPCA_
from ._base_datamining import BaseReduction
from ..core.decorators import validate_data_pca


class PCAError(Exception):
    """
    Custom exception for PCA class.
    """

    def __init__(self, message: str = "PCA error occurred."):
        self.message = message
        super().__init__(self.message)


class PCA(BaseReduction):
    """
    Principal Component Analysis (PCA) class.

    Attributes
    ----------
    n_components : int or float
        The number of components or the explained variance ratio.
    is_incremental : bool
        Indicates whether Incremental PCA is used.
    _pca : PCA_ or IncrementalPCA_
        The PCA or Incremental PCA model.
    is_fitted : bool
        Indicates whether the PCA model has been fitted.
    _data : xr.Dataset
        The original dataset.
    _stacked_data_matrix : np.ndarray
        The stacked data matrix.
    _standarized_stacked_data_matrix : np.ndarray
        The standardized stacked data matrix.
    scaler : StandardScaler
        The scaler used for standardizing the data.
    vars_to_stack : list of str
        The list of variables to stack.
    coords_to_stack : list of str
        The list of coordinates to stack.
    pca_dim_for_rows : str
        The dimension for rows in PCA.
    window_in_pca_dim_for_rows : list of int
        The window in PCA dimension for rows.
    value_to_replace_nans : float
        The value to replace NaNs in the dataset.
    num_cols_for_vars : int
        The number of columns for variables.

    Methods
    -------
    fit(
        data: xr.Dataset,
        vars_to_stack: List[str],
        coords_to_stack: List[str],
        pca_dim_for_rows: str,
        window_in_pca_dim_for_rows: List[int] = [0],
        value_to_replace_nans: float = None,
    ) -> None
    transform(data: xr.Dataset) -> xr.Dataset
    fit_transform(
        data: xr.Dataset,
        vars_to_stack: List[str],
        coords_to_stack: List[str],
        pca_dim_for_rows: str,
        window_in_pca_dim_for_rows: List[int] = [0],
        value_to_replace_nans: float = None,
    ) -> xr.Dataset
    inverse_transform(PCs: Union[np.ndarray, xr.Dataset]) -> xr.Dataset
    """

    def __init__(
        self, n_components: Union[int, float] = 0.98, is_incremental: bool = False
    ):
        """
        Initialize the PCA class.

        Parameters
        ----------
        n_components : int or float, optional
            Number of components to keep. If 0 < n_components < 1, it represents the
            proportion of variance to be explained by the selected components. If
            n_components >= 1, it represents the number of components to keep. Default is 0.98.
        is_incremental : bool, optional
            If True, use Incremental PCA which is useful for large datasets. Default is False.

        Raises
        ------
        ValueError
            If n_components is less than or equal to 0.
        TypeError
            If n_components is not an integer when it is greater than or equal to 1.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)
        if n_components <= 0:
            raise ValueError("Number of components must be greater than 0.")
        elif n_components >= 1:
            if not isinstance(n_components, int):
                raise TypeError("Number of components must be an integer when >= 1.")
            self.logger.info(f"Number of components: {n_components}")
        else:
            self.logger.info(f"Explained variance ratio: {n_components}")
        self.n_components = n_components
        if is_incremental:
            self.logger.info("Using Incremental PCA")
            self._pca = IncrementalPCA_(n_components=self.n_components)
            self.is_fitted: bool = False
        else:
            self.logger.info("Using PCA")
            self._pca = PCA_(n_components=self.n_components)
            self.is_fitted: bool = False
        self.is_incremental = is_incremental
        self._data: xr.Dataset = xr.Dataset()
        self._stacked_data_matrix: np.ndarray = np.array([])
        self._standarized_stacked_data_matrix: np.ndarray = np.array([])
        self.scaler: StandardScaler = StandardScaler()
        self.vars_to_stack: List[str] = []
        self.coords_to_stack: List[str] = []
        self.pca_dim_for_rows: str = None
        self.window_in_pca_dim_for_rows: List[int] = [0]
        self.value_to_replace_nans: float = None
        self.num_cols_for_vars: int = None

    @property
    def pca(self) -> Union[PCA_, IncrementalPCA_]:
        return self._pca

    @property
    def data(self) -> xr.Dataset:
        return self._data

    @property
    def stacked_data_matrix(self) -> np.ndarray:
        return self._stacked_data_matrix

    @property
    def standarized_stacked_data_matrix(self) -> np.ndarray:
        return self._standarized_stacked_data_matrix

    def _generate_stacked_data(self, data: xr.Dataset) -> np.ndarray:
        """
        Generate stacked data matrix.

        Parameters
        ----------
        data : xr.Dataset
            The data to stack.

        Returns
        -------
        stacked_data_matrix : np.ndarray
            The stacked data matrix
        """

        self.logger.info(
            f"Generating data matrix with variables to stack: {self.vars_to_stack} and coordinates to stack: {self.coords_to_stack}"
        )
        num_cols_for_vars = 1
        for coord_to_stack in self.coords_to_stack:
            num_cols_for_vars *= len(data[coord_to_stack])
        tmp_stacked_data = data.stack(positions=self.coords_to_stack)
        if (
            len(self.window_in_pca_dim_for_rows) != 0
            or self.window_in_pca_dim_for_rows[0] != 0
        ):
            self.logger.info(f"Rolling over coordinate: {self.pca_dim_for_rows}")
            tmp_stacked_data = xr.concat(
                [
                    tmp_stacked_data.shift({self.pca_dim_for_rows: i})
                    for i in self.window_in_pca_dim_for_rows
                ],
                dim="positions",
            )
        self.num_cols_for_vars = num_cols_for_vars * len(
            self.window_in_pca_dim_for_rows
        )
        stacked_data_matrix = np.hstack(
            [tmp_stacked_data[var].values for var in self.vars_to_stack]
        )
        self.logger.info(
            f"Data matrix generated successfully with shape: {self._stacked_data_matrix.shape}"
        )

        return stacked_data_matrix

    def _preprocess_data(self, data: xr.Dataset, is_fit: bool = True) -> np.ndarray:
        """
        Preprocess data for PCA. Steps:
        - Check NaNs and replace them if needed.
        - Generate stacked data matrix.
        - Standarize data matrix.
        - Remove NaNs from standarized data matrix.

        Parameters
        ----------
        data : xr.Dataset
            The data to preprocess.
        is_fit : bool, optional
            If True, set the data. Default is True.

        Returns
        -------
        standarized_stacked_data_matrix : np.ndarray
            The standarized stacked data matrix.
        """

        self.logger.info("Preprocessing data")
        data = self.check_nans(
            data=data,
            replace_value=self.value_to_replace_nans,
        )

        self.logger.info("Generating stacked data matrix")
        stacked_data_matrix = self._generate_stacked_data(
            data=data,
        )

        self.logger.info("Standarizing data matrix")
        standarized_stacked_data_matrix, scaler = self.standarize(
            data=stacked_data_matrix,
            scaler=self.scaler if not is_fit else None,
        )

        self.logger.info("Removing NaNs from standarized data matrix")
        standarized_stacked_data_matrix = self.check_nans(
            data=standarized_stacked_data_matrix,
            replace_value=self.value_to_replace_nans,
        )

        self.logger.info("Data preprocessed successfully")

        if is_fit:
            self._data = data.copy()
            self._stacked_data_matrix = stacked_data_matrix.copy()
            self._standarized_stacked_data_matrix = (
                standarized_stacked_data_matrix.copy()
            )
            self.scaler = scaler

        return standarized_stacked_data_matrix

    def _reshape_EOFs(self, destandarize: bool = False) -> xr.Dataset:
        """
        Reshape EOFs to the original data shape.

        Parameters
        ----------
        destandarize : bool, optional
            If True, destandarize the EOFs. Default is True.

        Returns
        -------
        xr.Dataset
            The reshaped EOFs.
        """

        EOFs = self.pca.components_  # Get Empirical Orthogonal Functions (EOFs)
        if destandarize:
            EOFs = self.scaler.inverse_transform(EOFs)
        EOFs_reshaped_vars_arrays = np.array_split(
            EOFs, len(self.vars_to_stack), axis=1
        )
        coords_to_stack_shape = [
            len(self.window_in_pca_dim_for_rows),
            self.pca.n_components_,
        ] + [self.data[coord].shape[0] for coord in self.coords_to_stack]
        EOFs_reshaped_vars_dict = {
            var: (
                ["window", "n_component", *self.coords_to_stack],
                np.array(
                    np.array_split(
                        EOF_reshaped_var, len(self.window_in_pca_dim_for_rows), axis=1
                    )
                ).reshape(*coords_to_stack_shape),
            )
            for var, EOF_reshaped_var in zip(
                self.vars_to_stack, EOFs_reshaped_vars_arrays
            )
        }
        return xr.Dataset(
            EOFs_reshaped_vars_dict,
            coords={
                "window": self.window_in_pca_dim_for_rows,
                "n_component": np.arange(self.pca.n_components_),
                **{coord: self.data[coord] for coord in self.coords_to_stack},
            },
        )

    def _reshape_data(self, X: np.ndarray, destandarize: bool = True) -> xr.Dataset:
        """
        Reshape data to the original data shape.

        Parameters
        ----------
        X : np.ndarray
            The data to reshape.
        destandarize : bool, optional
            If True, destandarize the data. Default is True.

        Returns
        -------
        xr.Dataset
            The reshaped data.
        """

        if destandarize:
            X = self.scaler.inverse_transform(X)
        X_reshaped_vars_arrays = np.array_split(X, len(self.vars_to_stack), axis=1)
        coords_to_stack_shape = [
            len(self.window_in_pca_dim_for_rows),
            self.data[self.pca_dim_for_rows].shape[0],
        ] + [self.data[coord].shape[0] for coord in self.coords_to_stack]
        X_reshaped_vars_dict = {
            var: (
                ["window", self.pca_dim_for_rows, *self.coords_to_stack],
                np.array(
                    np.array_split(
                        X_reshaped_var, len(self.window_in_pca_dim_for_rows), axis=1
                    )
                ).reshape(*coords_to_stack_shape),
            )
            for var, X_reshaped_var in zip(self.vars_to_stack, X_reshaped_vars_arrays)
        }
        return xr.Dataset(
            X_reshaped_vars_dict,
            coords={
                "window": self.window_in_pca_dim_for_rows,
                self.pca_dim_for_rows: self.data[self.pca_dim_for_rows],
                **{coord: self.data[coord] for coord in self.coords_to_stack},
            },
        )

    @validate_data_pca
    def fit(
        self,
        data: xr.Dataset,
        vars_to_stack: List[str],
        coords_to_stack: List[str],
        pca_dim_for_rows: str,
        window_in_pca_dim_for_rows: List[int] = [0],
        value_to_replace_nans: float = None,
    ) -> None:
        """
        Fit PCA model to data.

        Parameters
        ----------
        data : xr.Dataset
            The data to fit the PCA model.
        vars_to_stack : list of str
            The variables to stack.
        coords_to_stack : list of str
            The coordinates to stack.
        pca_dim_for_rows : str
            The PCA dimension to maintain in rows (usually the time).
        window_in_pca_dim_for_rows : list of int, optional
            The window steps to roll the pca_dim_for_rows. Default is [0].
        value_to_replace_nans : float, optional
            The value to replace NaNs. Default is None.
        """

        self.vars_to_stack = vars_to_stack
        self.coords_to_stack = coords_to_stack
        self.pca_dim_for_rows = pca_dim_for_rows
        self.window_in_pca_dim_for_rows = window_in_pca_dim_for_rows
        self.value_to_replace_nans = value_to_replace_nans

        self._preprocess_data(data=data[self.vars_to_stack], is_fit=True)
        self.logger.info("Fitting PCA model")
        self.pca.fit(X=self.standarized_stacked_data_matrix)
        self.is_fitted = True
        self.logger.info("PCA model fitted successfully")

    def transform(self, data: xr.Dataset) -> xr.Dataset:
        """
        Transform data using the fitted PCA model.

        Parameters
        ----------
        data : xr.Dataset
            The data to transform.

        Returns
        -------
        xr.Dataset
            The transformed data.
        """

        if self.is_fitted is False:
            raise PCAError("PCA model must be fitted before transforming data")
        self.logger.info("Transforming data using PCA model")
        processed_data = self._preprocess_data(
            data=data[self.vars_to_stack], is_fit=False
        )
        transformed_data = self.pca.transform(X=processed_data)
        return xr.Dataset(
            {
                "PCs": ((self.pca_dim_for_rows, "n_component"), transformed_data),
            },
            coords={
                self.pca_dim_for_rows: data[self.pca_dim_for_rows],
                "n_component": np.arange(self.pca.n_components_),
            },
        )

    def fit_transform(
        self,
        data: xr.Dataset,
        vars_to_stack: List[str],
        coords_to_stack: List[str],
        pca_dim_for_rows: str,
        window_in_pca_dim_for_rows: List[int] = [0],
        value_to_replace_nans: float = None,
    ) -> xr.Dataset:
        """
        Fit and transform data using PCA model.

        Parameters
        ----------
        data : xr.Dataset
            The data to fit the PCA model.
        vars_to_stack : list of str
            The variables to stack.
        coords_to_stack : list of str
            The coordinates to stack.
        pca_dim_for_rows : str
            The PCA dimension to maintain in rows (usually the time).
        window_in_pca_dim_for_rows : list of int, optional
            The window steps to roll the pca_dim_for_rows. Default is [0].
        value_to_replace_nans : float, optional
            The value to replace NaNs. Default is None.

        Returns
        -------
        xr.Dataset
            The transformed data.
        """

        self.fit(
            data=data,
            vars_to_stack=vars_to_stack,
            coords_to_stack=coords_to_stack,
            pca_dim_for_rows=pca_dim_for_rows,
            window_in_pca_dim_for_rows=window_in_pca_dim_for_rows,
            value_to_replace_nans=value_to_replace_nans,
        )
        # TODO: JAVI - Add a flag to use the already processed data??
        return self.transform(data=data)

    def inverse_transform(self, PCs: Union[np.ndarray, xr.Dataset]) -> xr.Dataset:
        """
        Inverse transform data using the fitted PCA model.

        Parameters
        ----------
        X : np.ndarray or xr.Dataset
            The data to inverse transform.

        Returns
        -------
        data_transformed : xr.Dataset
            The inverse transformed data.
        """

        if self.is_fitted is False:
            raise PCAError("PCA model must be fitted before inverse transforming data")
        if isinstance(PCs, xr.Dataset):
            X = PCs["PCs"].values
        elif isinstance(PCs, np.ndarray):
            X = PCs
        self.logger.info("Inverse transforming data using PCA model")
        X_transformed = self.pca.inverse_transform(X=X)
        data_transformed = self._reshape_data(X=X_transformed, destandarize=True)
        # Squeeze dataset and sort dimensions based on the original data
        data_transformed = data_transformed.squeeze().transpose(*self.data.dims)
        return data_transformed



from typing import List, Tuple
import numpy as np
import pandas as pd
from minisom import MiniSom
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
from ..core.plotting.base_plotting import DefaultStaticPlotting
from ..core.decorators import validate_data_som
from ._base_datamining import BaseClustering


class SOMError(Exception):
    """
    Custom exception for SOM class.
    """

    def __init__(self, message: str = "SOM error occurred."):
        self.message = message
        super().__init__(self.message)


class SOM(BaseClustering):
    """
    Self-Organizing Map (SOM) class.

    This class performs the Self-Organizing Map algorithm on a given dataframe.

    Attributes
    ----------
    som_shape : Tuple[int, int]
        The shape of the SOM.
    num_dimensions : int
        The number of dimensions of the input data.
    data : pd.DataFrame
        The input data.
    standarized_data : pd.DataFrame
        The standarized input data.
    data_to_fit : pd.DataFrame
        The data to fit the SOM algorithm.
    data_variables : List[str]
        A list with all data variables.
    directional_variables : List[str]
        A list with directional variables.
    fitting_variables : List[str]
        A list with fitting variables.
    scaler : StandardScaler
        The StandardScaler object.
    centroids : pd.DataFrame
        The selected centroids.
    is_fitted : bool
        A flag to check if the SOM model is fitted.

    Methods
    -------
    activation_response(data)
        Returns the activation response of the given data.
    get_centroids_probs_for_labels(data, labels)
        Returns the labels map of the given data.
    plot_centroids_probs_for_labels(probs_data)
        Plots the labels map of the given data.
    fit(data, directional_variables, num_iteration)
        Fits the SOM model to the provided data.
    predict(data)
        Predicts the nearest centroid for the provided data.
    fit_predict(data, directional_variables, num_iteration)
        Fit the SOM algorithm to the provided data and predict the nearest centroid for each data point.

    Notes
    -----
    - Check MiniSom documentation for more information:
        https://github.com/JustGlowing/minisom

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from bluemath_tk.datamining.som import SOM
    >>> data = pd.DataFrame(
    ...     {
    ...         'Hs': np.random.rand(1000) * 7,
    ...         'Tp': np.random.rand(1000) * 20,
    ...         'Dir': np.random.rand(1000) * 360
    ...     }
    ... )
    >>> som = SOM(som_shape=(3, 3), num_dimensions=4)
    >>> nearest_centroids_idxs, nearest_centroids_df = som.fit_predict(
    ...     data=data,
    ...     directional_variables=['Dir'],
    ... )

    TODO
    -----
    - Add option to normalize data?
    """

    def __init__(
        self,
        som_shape: Tuple[int, int],
        num_dimensions: int,
        sigma: float = 1,
        learning_rate: float = 0.5,
        decay_function: str = "asymptotic_decay",
        neighborhood_function: str = "gaussian",
        topology: str = "rectangular",
        activation_distance: str = "euclidean",
        random_seed: int = None,
        sigma_decay_function: str = "asymptotic_decay",
    ) -> None:
        """
        Initializes a Self Organizing Maps.

        A rule of thumb to set the size of the grid for a dimensionality
        reduction task is that it should contain 5*sqrt(N) neurons
        where N is the number of samples in the dataset to analyze.

        E.g. if your dataset has 150 samples, 5*sqrt(150) = 61.23
        hence a map 8-by-8 should perform well.

        Parameters
        ----------
        som_shape : tuple
            Shape of the SOM. This should be a tuple with two integers.
        num_dimensions : int
            Number of the elements of the vectors in input.

        For the other parameters, check the MiniSom documentation:
            https://github.com/JustGlowing/minisom/blob/master/minisom.py

        Raises
        ------
        ValueError
            If the SOM shape is not a tuple with two integers.
            Or if the number of dimensions is not an integer.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)
        if not isinstance(som_shape, tuple):
            if len(som_shape) != 2:
                raise ValueError("Invalid SOM shape.")
        self.som_shape = som_shape
        if not isinstance(num_dimensions, int):
            raise ValueError("Invalid number of dimensions.")
        self.num_dimensions = num_dimensions
        self.x = self.som_shape[0]
        self.y = self.som_shape[1]
        self.sigma = sigma
        self.learning_rate = learning_rate
        self.decay_function = decay_function
        self.neighborhood_function = neighborhood_function
        self.topology = topology
        self.activation_distance = activation_distance
        self.random_seed = random_seed
        self.sigma_decay_function = sigma_decay_function
        self._som = MiniSom(
            x=self.x,
            y=self.y,
            input_len=self.num_dimensions,
            sigma=self.sigma,
            learning_rate=self.learning_rate,
            decay_function=self.decay_function,
            neighborhood_function=self.neighborhood_function,
            topology=self.topology,
            activation_distance=self.activation_distance,
            random_seed=self.random_seed,
            sigma_decay_function=self.sigma_decay_function,
        )
        self._data: pd.DataFrame = pd.DataFrame()
        self._standarized_data: pd.DataFrame = pd.DataFrame()
        self._data_to_fit: pd.DataFrame = pd.DataFrame()
        self.data_variables: List[str] = []
        self.directional_variables: List[str] = []
        self.fitting_variables: List[str] = []
        self.scaler: StandardScaler = StandardScaler()
        self.centroids: pd.DataFrame = pd.DataFrame()
        self.is_fitted: bool = False

    @property
    def som(self) -> MiniSom:
        return self._som

    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @property
    def standarized_data(self) -> pd.DataFrame:
        return self._standarized_data

    @property
    def data_to_fit(self) -> pd.DataFrame:
        return self._data_to_fit

    @property
    def distance_map(self) -> np.ndarray:
        """
        Returns the distance map of the SOM.
        """

        return self.som.distance_map().T

    def _get_winner_neurons(self, standarized_data: np.ndarray) -> np.ndarray:
        """
        Returns the winner neurons of the given standarized data.
        """

        winner_neurons = np.array([self.som.winner(x) for x in standarized_data]).T
        return np.ravel_multi_index(winner_neurons, self.som_shape)

    def activation_response(self, data: pd.DataFrame = None) -> np.ndarray:
        """
        Returns the activation response of the given data.
        """

        if data is None:
            data = self.standarized_data.copy()
        else:
            data, _ = self.standarize(data=data, scaler=self.scaler)

        return self.som.activation_response(data=data.values)

    def get_centroids_probs_for_labels(
        self, data: pd.DataFrame, labels: List[str]
    ) -> pd.DataFrame:
        """
        Returns the labels map of the given data.
        """

        # TODO: JAVI: Could this method be implemented in more datamining classes?
        data = data.copy()  # Avoid modifying the original data to predict
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=data[directional_variable].values
            )
            data[f"{directional_variable}_u"] = u_comp
            data[f"{directional_variable}_v"] = v_comp
            data.drop(columns=[directional_variable], inplace=True)
        standarized_data, _ = self.standarize(data=data, scaler=self.scaler)
        dict_with_probs = self.som.labels_map(standarized_data.values, labels)
        return pd.DataFrame(dict_with_probs).T.sort_index()

    def plot_centroids_probs_for_labels(
        self, probs_data: pd.DataFrame
    ) -> Tuple[plt.figure, plt.axes]:
        """
        Plots the labels map of the given data.
        """

        default_static_plot = DefaultStaticPlotting()
        fig, axes = default_static_plot.get_subplots(
            nrows=self.som_shape[0],
            ncols=self.som_shape[1],
        )
        for index in probs_data.index:
            default_static_plot.plot_pie(
                ax=axes[*index], x=probs_data.loc[index], labels=probs_data.columns
            )

        return fig, axes

    @validate_data_som
    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        num_iteration: int = 1000,
    ) -> None:
        """
        Fits the SOM model to the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the fitting.
        directional_variables : List[str], optional
            A list with the directional variables (will be transformed to u and v).
            Default is [].
        num_iteration : int, optional
            The number of iterations for the SOM fitting.
            Default is 1000.

        Notes
        -----
        - The function assumes that the data is validated by the `validate_data_som`
        decorator before execution.
        """

        self._data = data.copy()
        self.directional_variables = directional_variables.copy()
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=self.data[directional_variable].values
            )
            self.data[f"{directional_variable}_u"] = u_comp
            self.data[f"{directional_variable}_v"] = v_comp
        self.data_variables = list(self.data.columns)

        # Get just the data to be used in the training
        self._data_to_fit = self.data.copy()
        for directional_variable in self.directional_variables:
            self.data_to_fit.drop(columns=[directional_variable], inplace=True)
        self.fitting_variables = list(self.data_to_fit.columns)

        # Standarize data using the StandardScaler custom method
        self._standarized_data, self.scaler = self.standarize(data=self.data_to_fit)

        # Train the SOM model
        self.som.train(data=self.standarized_data.values, num_iteration=num_iteration)

        # Save winner neurons and calculate centroids values
        data_and_winners = self.data.copy()
        data_and_winners["winner_neurons"] = self._get_winner_neurons(
            standarized_data=self.standarized_data.values
        )
        self.centroids = data_and_winners.groupby("winner_neurons").mean()
        for directional_variable in self.directional_variables:
            self.centroids[directional_variable] = self.get_degrees_from_uv(
                xu=self.centroids[f"{directional_variable}_u"].values,
                xv=self.centroids[f"{directional_variable}_v"].values,
            )

        # Set the fitted flag to True
        self.is_fitted = True

    def predict(self, data: pd.DataFrame) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Predicts the nearest centroid for the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the prediction.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple with the winner neurons and the centroids of the given data.
        """

        if self.is_fitted is False:
            raise SOMError("SOM model is not fitted.")
        data = data.copy()  # Avoid modifying the original data to predict
        for directional_variable in self.directional_variables:
            u_comp, v_comp = self.get_uv_components(
                x_deg=data[directional_variable].values
            )
            data[f"{directional_variable}_u"] = u_comp
            data[f"{directional_variable}_v"] = v_comp
            data.drop(columns=[directional_variable], inplace=True)
        standarized_data, _ = self.standarize(data=data, scaler=self.scaler)
        winner_neurons = self._get_winner_neurons(
            standarized_data=standarized_data.values
        )

        return winner_neurons, self.centroids.iloc[winner_neurons]

    def fit_predict(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        num_iteration: int = 1000,
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Fit the SOM algorithm to the provided data and predict the nearest centroid for each data point.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the SOM algorithm.
        directional_variables : List[str], optional
            A list of directional variables (will be transformed to u and v).
            Default is [].
        num_iteration : int, optional
            The number of iterations for the SOM fitting.
            Default is 1000.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple containing the winner neurons for each data point and the nearest centroids.
        """

        self.fit(
            data=data,
            directional_variables=directional_variables,
            num_iteration=num_iteration,
        )

        return self.predict(data=data)
